#! /bin/jsh -
### ====================================================================
### Search the Library of Congress catalog via its Z39.50 interface for
### one or more publications by author, title, ISBN, or ISSN, and
### convert the US MARC records found to BibTeX on stdout.  It is NOT
### necessary to remove hyphens from ISBN or ISSN data.
###
### Usage:
###	cattobib [ --? ] \
###		 [ --author ] \
###		 [ --byxxx ] \
###		 [ --CODEN ] \
###		 [ --debug ] \
###		 [ --editor ] \
###		 [ --help ] \
###		 [ --ISBN ] \
###		 [ --ISSN ] \
###		 [ --keep-files ] \
###		 [ --logfile new-filename ] \
###		 [ --quiet ] \
###		 [ --server server-address ] \
###		 [ --test ] \
###		 [ --title ] \
###		 [ --version ] \
###		 search-key-1 search-key-2 search-key-3 ... >BibTeX-file
###
### Options may be abbreviated to any unique leading prefix, and may
### begin with either one or two hyphens.  Any option beginning -by...
### is assumed to be a bibsort option, and is passed verbatim to it.
###
### [15-Nov-2004]
### ====================================================================

IFS='
 	'
umask 077


### Root of local installation tree
prefix=/usr/local
bindir=$prefix/bin

TMPDIR=${TMPDIR-/tmp}

### Paths and directories:
PROGRAM=cattobib
VERSION=0.03
SHRLIBDIR=${TESTSHRLIBDIR-$prefix/share/$PROGRAM/${PROGRAM}-$VERSION}

### Programs needed:
AWK=/usr/local/bin/nawk
BIBCLEAN=/usr/local/bin/bibclean
BIBLABEL=/usr/local/bin/biblabel
BIBORDER=/usr/local/bin/biborder
BIBORDERFLAGS=--isbn-13
BIBSORT=/usr/local/bin/bibsort
CAT=/usr/local/bin/cat
CITESUB=/usr/local/bin/citesub
DATE=/usr/local/bin/date
EXPECT=/usr/local/bin/expect
HOSTNAME=`/usr/local/bin/hostname`
LOGGER=/usr/local/bin/tee
SED=/usr/local/bin/sed
TEE=/usr/local/bin/tee
TEST=/bin/test
YAZ_CLIENT=/usr/local/bin/yaz-client
Z39_50_TO_BIBTEX=$SHRLIBDIR/usmarc-to-bibtex.awk

### Files, flags and variables:
BIBCLEANFLAGS="-max-width 0"
BIBCLEANFLAGS=
BIBSORTFLAGS=
DEBUG=0
DEFAULT_Z39_50_SERVER="z3950.loc.gov:7090/Voyager"
FIFO=$TMPDIR/$PROGRAM.fifo.$$
flags=
FORMAT="USMARC"
LABELFILE=$TMPDIR/$PROGRAM.label.$$
LOGFILE=$TMPDIR/$PROGRAM.$HOSTNAME.`$DATE +%Y.%m.%d.%H.%M.%S`
LOGFILE=/dev/null
PACKAGE_DATE="22-Dec-2005"
PACKAGE_VERSION="0.03"
SEDINPUTFILTER=$CAT
SEDOUTPUTFILTER=$CAT
TEST_MODE=no
TMPBIBTEXFILE=$TMPDIR/$PROGRAM.bib.$$

# Z39.50 catalog servers:
Z39_50_SERVER="$DEFAULT_Z39_50_SERVER"
Z39_50_SERVER_LIST=""

# Ensure consistent sorting across all platforms
LC_ALL=C
export LC_ALL

trap "rm -f $FIFO $LABELFILE $TMPBIBTEXFILE"  EXIT HUP INT QUIT PIPE TERM

error_and_exit()
{
    echo "ERROR: ""$@" 1>&2
    exit 1
}

usage()
{
    cat <<EOF
$0 [--?]
		    [ --author ]
		    [ --byxxx ]
		    [ --CODEN ]
		    [ --debug ]
		    [ --editor ]
		    [ --help ]
		    [ --ISBN ]
		    [ --ISSN ]
		    [ --keep-files ]
		    [ --logfile new-filename ]
		    [ --quiet ]
		    [ --server server-address ]
		    [ --test ]
		    [ --title ]
		    [ --version ]
		    search-key-1 search-key-2 search-key-3 ... >BibTeX-file

Search keys should be either ISBNs, or quoted title strings, unless
one of qualifier options --author, --CODEN, --editor, --ISBN, --ISSN,
or --title is specified, in which case, the search strings should be
of that type.

All searches are completed before the data is processed further to
produce BibTeX entries, so if you interrupt execution, there will
likely be no useful output.

The server can be supplied as an explicit Z39.50 address, such as

			 $DEFAULT_Z39_50_SERVER

(the default server at the US Library of Congress), or more usually, by
a shorthand name (vertical bars separate alternatives, and asterisk
matches any word with that prefix):

	ALL | A*			All Z39.50 servers known to cattobib
	NATIONAL | N*			National libraries and union catalogs
	alberta | ab			University of Alberta
	amherst | umass			University of Massachusetts, Amherst
	amicus | ca			National Library of Canada
	bibsys              		Norwegian Union Catalog
	bne | es			National Library of Spain
	bnp | pt            		National Library of Portugal
	boulder | co			University of Colorado, Boulder
	british | br			British Library
	calgary				University of Calgary
	columbia | cu			Columbia University
	congress | lc			US Library of Congress
	copac | uk			COPAC (union of 24 research-university
					catalogs in the UK and Ireland)
	denmark | dk			Royal Library of Denmark
	dsb | dsl			Danish State Library
	duke				Duke University
	florida | fl			Florida Center for Library Automation
	gbv				German Union Catalog
	marriott | ut			University of Utah Marriott Library
	melbourne			University of Melbourne
	melvyl | cal			University of California MELVYL catalog
	minn* | mn			University of Minnesota
	mit				Massachusetts Institute of Technology
	newyork | ny			New York University
	nla | au			National Library of Australia
	nlm | nih			National Library of Medicine (US)
	nlnz | nz			National Library of New Zealand
	nls				National Library of Scotland
	norway | no			National Library of Norway
	nsw | unsw			University of New South Wales
	odense | sdu			University of Southern Denmark
	oxford | ox*			Oxford University
	poland | pl			National Library of Poland
	rlg				Research Libraries Group
	sudoc | abes | fr		French Union Catalog
	sweden | se			National Library of Sweden
	texas | tx			University of Texas at Austin
	toronto				University of Toronto
	usc				University of Southern California

Please note that data from ANY of these libraries are likely to be error
ridden: it is best to repeat your searches with three or more servers in
order to improve data quality and reliability.

EOF
}

usage_and_exit()
{
    usage
    exit 0
}

version_and_exit()
{
    echo "This is `basename $0` version $PACKAGE_VERSION [$PACKAGE_DATE]"
    exit 0
}

while test $# -gt 0
do
	case $1 in

	'--?' | '-?' )
		usage_and_exit
		;;

	--author | --autho | --auth | --aut | --au | --a | \
	-author | -autho | -auth | -aut | -au | -a )
		flags="@attr 1=1003"
		flags="@attr 1=1"
		;;

	--by* | -by*)
		shift
		BIBSORT=$bindir/bibsort
		BIBSORTFLAGS="$BIBSORTFLAGS $1"
		;;

	--coden | --code | --cod | --co | --c | \
	-coden | -code | -cod | -co | -c | \
	--CODEN | --CODE | --COD | --CO | --C | \
	-CODEN | -CODE | -COD | -CO | -C )
		flags="@attr 1=60"
		;;

	--debug | --debu | --deb | --de | --d | \
	-debug | -debu | -deb | -de | -d )
		DEBUG=1
		;;

	--editor | --edito | --edit | --edi | --ed | --e | \
	-editor | -edito | -edit | -edi | -ed | -e )
		flags="@attr 1=1020"
		;;

	--help | --hel | --he | --h | \
	-help | -hel | -he | -h )
		usage_and_exit
		;;

	--isbn | --isb | -isbn | -isb | -is | -i | \
	--ISBN | --ISB | -ISBN | -ISB | -IS | -I )
		flags="@attr 1=7"
		;;

	--issn | --iss | -issn | -iss | \
	--ISSN | --ISS | -ISSN | -ISS )
		flags="@attr 1=8"
		;;

	--keep-files | --keep-file | --keep-fil | --keep-fi | --keep-f | --keep- | \
	--keep | --kee | --ke | --k | \
	-keep-files | -keep-file | -keep-fil | -keep-fi | -keep-f | -keep- | \
	-keep | -kee | -ke | -k )
		trap '' EXIT HUP INT QUIT PIPE TERM
		;;

	--log | --lo | --l | \
	-log | -lo | -l )
		shift
		$TEST -f "$LOGFILE" && error_and_exit "Log file $LOGFILE already exists"
		;;

	--quiet | --quie | --qui | --qu | --q | \
	-quiet | -quie | -qui | -qu | -q )
		LOGGER=$CAT
		LOGFILE=
		;;

	--server | --serve | --serv | --ser | --se | --s | \
	-server | -serve | -serv | -ser | -se | -s )
		shift
		## See http://www.indexdata.dk/targettest/u.php for a long list
		## See http://www.bs.dk/danzig for pointers to a list of Danish servers
		## See http://www.nla.gov.au/apps/libraries?action=ListTargets for Australian servers

		for z in $1
		do
			case $z in

			### TO DO: National Library of Portugal is SUTRS
			### format, and not properly supported by ALL and
			### NATIONAL server requests.

			### These work
			ALL | A* )				Z39_50_SERVER="dra.library.ualberta.ca:210/Marion
										umlibr.library.umass.edu:210/INNOPAC
										amicus.nlc-bnc.ca:210/NL
										z3950.bibsys.no:2100/BIBSYS
										www.bne.es:2210/BIMO
										malvine.bn.pt:2100
										libraries.colorado.edu:210/INNOPAC
										z3950cat.bl.uk:9909/BLAC
										clavis.ucalgary.ca:2200/UNICORN
										clio-db.cc.columbia.edu:7090/Voyager
										z3950.loc.gov:7090/Voyager
										z3950.copac.ac.uk:2020
										z3950.kb.dk:2100/KGL01
										zeus.statsbiblioteket.dk:210/Horizon
										ducatalog.lib.duke.edu:210/MARION
										z3950.fcla.edu:210/AM
										z3950.gbv.de:20011/gvk
										cos.lib.utah.edu:6060
										library.unimelb.edu.au:210/INNOPAC
										melvyl.cdlib.org:210/CDL90
										umnlib.oit.umn.edu:9909/MNCAT
										bobcat.nyu.edu:210/ADVANCE
										catalogue.nla.gov.au:7090/Voyager
										tegument.nlm.nih.gov:7090/Voyager
										zed.natlib.govt.nz:210/pinz
										z3950.nls.uk:7290/voyager
										z3950.nb.no:2100/Norper
										aleph.nun.unsw.edu.au:80/NUN
										odin2.bib.sdu.dk:210/Horizon
										library.ox.ac.uk:210/ADVANCE
										alpha.bn.org.pl:210/INNOPAC
										zinc.rlg.org:210/SAM
										carmin.sudoc.abes.fr:210/ABES-Z39-PUBLIC
										z3950.libris.kb.se:210/libr
										db.lib.utexas.edu:210/utcat
										sirsi.library.utoronto.ca:2200/UNICORN
										library.usc.edu:2200/unicorn"
								;;
			NATIONAL | N* )				Z39_50_SERVER="amicus.nlc-bnc.ca:210/NL
										z3950.bibsys.no:2100/BIBSYS
										www.bne.es:2210/BIMO
										malvine.bn.pt:2100
										z3950cat.bl.uk:9909/BLAC
										z3950.loc.gov:7090/Voyager
										z3950.copac.ac.uk:2020
										z3950.kb.dk:2100/KGL01
										zeus.statsbiblioteket.dk:210/Horizon
										z3950.gbv.de:20011/gvk
										catalogue.nla.gov.au:7090/Voyager
										tegument.nlm.nih.gov:7090/Voyager
										zed.natlib.govt.nz:210/pinz
										z3950.nls.uk:7290/voyager
										z3950.nb.no:2100/Norper
										alpha.bn.org.pl:210/INNOPAC
										carmin.sudoc.abes.fr:210/ABES-Z39-PUBLIC
										z3950.libris.kb.se:210/libr
										sirsi.library.utoronto.ca:2200/UNICORN"
								;;
			alberta | ab )				Z39_50_SERVER="dra.library.ualberta.ca:210/Marion" ;;
			amherst | umass )			Z39_50_SERVER="umlibr.library.umass.edu:210/INNOPAC" ;;
			amicus | ca )				Z39_50_SERVER="amicus.nlc-bnc.ca:210/NL" ;;
			bibsys )				Z39_50_SERVER="z3950.bibsys.no:2100/BIBSYS" ;;
			bne | es )				Z39_50_SERVER="www.bne.es:2210/BIMO" ;;
			bnp | pt )				Z39_50_SERVER="malvine.bn.pt:2100"
								Z39_50_TO_BIBTEX="$SHRLIBDIR/sutrs-to-bibtex.awk"
								FORMAT="SUTRS"
								;;
			boulder | co )				Z39_50_SERVER="libraries.colorado.edu:210/INNOPAC" ;;
			british | br )				Z39_50_SERVER="z3950cat.bl.uk:9909/BLAC"
								Z39_50_TO_BIBTEX="$SHRLIBDIR/sutrs-to-bibtex.awk"
								FORMAT="SUTRS"
								URL="http://www.bl.uk/catalogues/z3950copacaccess.html"
								;;
			calgary )				Z39_50_SERVER="clavis.ucalgary.ca:2200/UNICORN" ;;
			columbia | cu )				Z39_50_SERVER="clio-db.cc.columbia.edu:7090/Voyager" ;;
			congress | lc )				Z39_50_SERVER="z3950.loc.gov:7090/Voyager" ;;
			copac | uk )				Z39_50_SERVER="z3950.copac.ac.uk:2020" ;;
			denmark | dk )				Z39_50_SERVER="z3950.kb.dk:2100/KGL01" ;;
			dsb | dsl )				Z39_50_SERVER="zeus.statsbiblioteket.dk:210/Horizon" ;;
			duke )					Z39_50_SERVER="ducatalog.lib.duke.edu:210/MARION" ;;
			florida | fl )				Z39_50_SERVER="z3950.fcla.edu:210/AM" ;;
			gbv )					Z39_50_SERVER="z3950.gbv.de:20011/gvk" ;;
			marriott | ut )				Z39_50_SERVER="cos.lib.utah.edu:6060" ;;
			melbourne )				Z39_50_SERVER="library.unimelb.edu.au:210/INNOPAC" ;;
			melvyl | cal )				Z39_50_SERVER="melvyl.cdlib.org:210/CDL90" ;;
			minn* | mn )				Z39_50_SERVER="umnlib.oit.umn.edu:9909/MNCAT" ;;
			mit )					Z39_50_SERVER="library.mit.edu:9909/mit01" ;;
			newyork | ny )				Z39_50_SERVER="bobcat.nyu.edu:210/ADVANCE" ;;
			nla | au )				Z39_50_SERVER="catalogue.nla.gov.au:7090/Voyager" ;;
			nlm | nih )				Z39_50_SERVER="tegument.nlm.nih.gov:7090/Voyager" ;;
			nlnz | nz )				Z39_50_SERVER="zed.natlib.govt.nz:210/pinz" ;;
			nls )					Z39_50_SERVER="z3950.nls.uk:7290/voyager" ;;
			norway | no )				Z39_50_SERVER="z3950.nb.no:2100/Norper" ;;
			nsw | unsw )				Z39_50_SERVER="aleph.nun.unsw.edu.au:80/NUN" ;;
			odense | sdu )				Z39_50_SERVER="odin2.bib.sdu.dk:210/Horizon" ;;
			oxford | ox* )				Z39_50_SERVER="library.ox.ac.uk:210/ADVANCE" ;;
			poland | pl )				Z39_50_SERVER="alpha.bn.org.pl:210/INNOPAC"
								SEDINPUTFILTER="$SED -f $SHRLIBDIR/TeX-to-Polish.sed"
								SEDOUTPUTFILTER="$SED -f $SHRLIBDIR/Polish-to-TeX.sed"
								;;
			rlg )					Z39_50_SERVER="zinc.rlg.org:210/SAM" ;;
			sudoc | abes | fr )			Z39_50_SERVER="carmin.sudoc.abes.fr:210/ABES-Z39-PUBLIC" ;;
			sweden | se )				Z39_50_SERVER="z3950.libris.kb.se:210/libr" ;;
			texas | tx )				Z39_50_SERVER="db.lib.utexas.edu:210/utcat" ;;
			toronto )				Z39_50_SERVER="sirsi.library.utoronto.ca:2200/UNICORN" ;;
			usc )					Z39_50_SERVER=library.usc.edu:2200/unicorn ;;

			### These do not work
			## bell* | z3950.bell-labs.com )	Z39_50_SERVER=z3950.bell-labs.com/books ;;
			## inquirion* | teratext* )		Z39_50_SERVER="z3950.inquirion.com:210" ;;
			## melvyl* | calif* )			Z39_50_SERVER=melvyl.ucop.edu/CAT ;;

			### Default: user-provided address
			*)					Z39_50_SERVER="$z" ;;
			esac
			Z39_50_SERVER_LIST="$Z39_50_SERVER_LIST $Z39_50_SERVER"
		done
		;;

	--test | --tes | --te | \
	-test | -tes | -te )
		TEST_MODE=yes
		;;

	--title | --titl | --tit | --ti | --t | \
	-title | -titl | -tit | -ti | -t )
		flags="@attr 1=4"
		;;

	--version | --versio | --versi | --vers | --ver | --ve | --v | \
	-version | -versio | -versi | -vers | -ver | -ve | -v )
		version_and_exit
		;;

	-*)
		error_and_exit unrecognized option "$1"
		;;

	*)
		break
		;;
	esac
	shift
done

test $LOGGER = $CAT && LOGFILE=
test "$Z39_50_SERVER_LIST" = "" && Z39_50_SERVER_LIST="$DEFAULT_Z39_50_SERVER"

# mkfifo $FIFO
for Z39_50_SERVER in $Z39_50_SERVER_LIST
do
	for f in "$@"
	do
		case $f in
		[0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9Xx] | \
		[-0-9][-0-9][-0-9][-0-9][-0-9][-0-9][-0-9][-0-9][-0-9][-0-9][-0-9][-0-9][-0-9Xx] | \
		[0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9] | \
		[-0-9][-0-9][-0-9][-0-9][-0-9][-0-9][-0-9][-0-9][-0-9][-0-9][-0-9][-0-9][-0-9][-0-9][-0-9][-0-9][-0-9])
			# Remove hyphens from ISBNs (old 10-digit and new 13-digit)
			f=`echo $f | sed -e s/-//g`
			flags="@attr 1=7"
			;;
		*)
			f=`echo $f | $SEDINPUTFILTER`
			;;
		esac
		test $LOGGER = $CAT || echo "%% Searching [$Z39_50_SERVER] for [$f]: flags = [$flags]" 1>&2

		if test "$TEST_MODE" = yes
		then			# test input from stdin instead of from Z39.50 server
			$CAT >$FIFO
		else
			if test $DEBUG = 1
			then
				$CAT 1>&2 <<-EOFILE

					%% BEGIN DEBUG
					$YAZ_CLIENT
					open $Z39_50_SERVER
					format $FORMAT
					find $flags "$f"
					show
					...
					show
					close
					%% END DEBUG

EOFILE
			fi
			$EXPECT >$FIFO <<-EOFILE
				spawn $YAZ_CLIENT
				expect "Z>"
				send -- "open $Z39_50_SERVER\r"
				expect "Z>"
				send -- "format $FORMAT\r"
				expect "Z>"
				send -- "find $flags \"$f\"\r"
				expect "Number of hits: "
				expect "Z>"
				# For now, just try to display first twenty-five records
				send -- "show\r"
				expect "Z>"
				send -- "show\r"
				expect "Z>"
				send -- "show\r"
				expect "Z>"
				send -- "show\r"
				expect "Z>"
				send -- "show\r"
				expect "Z>"
				send -- "show\r"
				expect "Z>"
				send -- "show\r"
				expect "Z>"
				send -- "show\r"
				expect "Z>"
				send -- "show\r"
				expect "Z>"
				send -- "show\r"
				expect "Z>"
				send -- "show\r"
				expect "Z>"
				send -- "show\r"
				expect "Z>"
				send -- "show\r"
				expect "Z>"
				send -- "show\r"
				expect "Z>"
				send -- "show\r"
				expect "Z>"
				send -- "show\r"
				expect "Z>"
				send -- "show\r"
				expect "Z>"
				send -- "show\r"
				expect "Z>"
				send -- "show\r"
				expect "Z>"
				send -- "show\r"
				expect "Z>"
				send -- "show\r"
				expect "Z>"
				send -- "show\r"
				expect "Z>"
				send -- "show\r"
				expect "Z>"
				send -- "show\r"
				expect "Z>"
				send -- "show\r"
				send -- "close\r"
EOFILE
		fi
		if test $DEBUG = 1
		then
			## Modern FreeBSD, GNU/Linux, Mac OS X, NetBSD, OpenBSD,
			## and Solaris have both /dev/stderr and /dev/fd/2
			## accessible to programs (not just certain shells).
			## IRIX has only /dev/fd/2.  Alas, HP-UX and OSF/1 have
			## neither.  Because test is often a shell builtin, and
			## some shells on those systems fake those devices, we
			## need to avoid the built-in version: TEST is defined
			## above as /bin/test instead.
			if $TEST -w /dev/stderr
			then
				debugfile=/dev/stderr
			elif $TEST -w /dev/fd/2
			then
				debugfile=/dev/fd/2
			else
				debugfile=$TMPDIR/$PROGRAM.debug.$$
				$TEST -f $debugfile || echo "%% DEBUG output logged to $debugfile"
			fi
			DEBUGFILTER="$TEE -a $debugfile"
		else
			DEBUGFILTER="$CAT"
		fi
		cat $FIFO |
tee foo.tee.1 |
			eval $DEBUGFILTER |
tee foo.tee.2 |
				eval $SEDOUTPUTFILTER |
tee foo.tee.3 |
					$AWK -f $Z39_50_TO_BIBTEX
	done  > $TMPBIBTEXFILE
	$BIBLABEL $TMPBIBTEXFILE >$LABELFILE

cat $TMPBIBTEXFILE > foo.tee.4


	$CITESUB -f $LABELFILE $TMPBIBTEXFILE 2>/dev/null |
tee foo.tee.5 |
		 $BIBSORT $BIBSORTFLAGS |
tee foo.tee.6 |
			 $BIBCLEAN $BIBCLEANFLAGS 2>/dev/null |
tee foo.tee.7 |
				 $BIBORDER $BIBORDERFLAGS |
tee foo.tee.8 |
					$AWK -f $SHRLIBDIR/iso8859-1-to-TeX.awk |
tee foo.tee.9 |
						$LOGGER $LOGFILE
done
